mod http;
mod udp;
mod usimp;
mod websocket;

use openssl::ssl::{SslAcceptor, SslFiletype, SslMethod, SslStream};
use std::net::{TcpListener, UdpSocket, SocketAddr};
use std::sync::Arc;
use std::sync::Mutex;
use std::thread;
use threadpool::ThreadPool;
use r2d2_postgres::{PostgresConnectionManager, postgres::NoTls};
use r2d2;
use r2d2::{ManageConnection, Pool};
use std::ops::Deref;

enum SocketType {
    Http,
    Https,
    Udp,
}

struct SocketConfig {
    address: SocketAddr,
    socket_type: SocketType,
}

pub enum BackendPool {
    Postgres(Pool<PostgresConnectionManager<NoTls>>),
}

pub enum BackendClient {
    Postgres(r2d2::PooledConnection<PostgresConnectionManager<NoTls>>),
}

static mut DB_POOL: Option<Arc<Mutex<BackendPool>>> = None;

pub fn get_backend() -> BackendClient {
    unsafe {
        match DB_POOL.as_ref().unwrap().clone().lock().unwrap().deref() {
            BackendPool::Postgres(pool) => {
                BackendClient::Postgres(pool.get().unwrap())
            }
        }
    }
}

fn main() {
    let socket_configs: Vec<SocketConfig> = vec![
        SocketConfig {
            address: "[::]:8080".parse().unwrap(),
            socket_type: SocketType::Http,
        },
        SocketConfig {
            address: "[::]:8443".parse().unwrap(),
            socket_type: SocketType::Https,
        },
        SocketConfig {
            address: "[::]:3126".parse().unwrap(),
            socket_type: SocketType::Udp,
        }
    ];

    let db_manager = PostgresConnectionManager::new(
        "host=localhost user=postgres dbname=test".parse().unwrap(),
        NoTls,
    );
    let db_pool = r2d2::Pool::new(db_manager).unwrap();
    unsafe {
        DB_POOL = Some(Arc::new(Mutex::new(BackendPool::Postgres(db_pool))));
    }

    let thread_pool = ThreadPool::new(256);
    let thread_pool_mutex = Arc::new(Mutex::new(thread_pool));

    let mut threads = Vec::new();

    for socket_config in socket_configs {
        let thread_pool_mutex = thread_pool_mutex.clone();

        threads.push(match socket_config.socket_type {
            SocketType::Http => thread::spawn(move || {
                let mut tcp_socket = TcpListener::bind(socket_config.address).unwrap();

                for stream in tcp_socket.incoming() {
                    thread_pool_mutex.lock().unwrap().execute(|| {
                        let stream = stream.unwrap();
                        http::connection_handler(http::Stream::Tcp(stream));
                    });
                }
            }),
            SocketType::Https => thread::spawn(move || {
                let mut ssl_socket = TcpListener::bind(socket_config.address).unwrap();

                let mut acceptor = SslAcceptor::mozilla_intermediate(SslMethod::tls()).unwrap();
                acceptor
                    .set_certificate_chain_file("/home/lorenz/Certificates/chakotay.pem")
                    .unwrap();
                acceptor
                    .set_private_key_file(
                        "/home/lorenz/Certificates/priv/chakotay.key",
                        SslFiletype::PEM,
                    )
                    .unwrap();
                acceptor.check_private_key().unwrap();
                let acceptor = Arc::new(acceptor.build());

                for stream in ssl_socket.incoming() {
                    let acceptor = acceptor.clone();
                    thread_pool_mutex.lock().unwrap().execute(move || {
                        let stream = stream.unwrap();
                        let stream = acceptor.accept(stream).unwrap();
                        http::connection_handler(http::Stream::Ssl(stream));
                    });
                }
            }),
            SocketType::Udp => thread::spawn(move || {
                let mut udp_socket = UdpSocket::bind(socket_config.address).unwrap();
                let mut buf = [0; 65_536];

                loop {
                    let (size, addr) = udp_socket.recv_from(&mut buf).unwrap();
                    let req = udp::Request::new(&udp_socket, addr, size, &buf);
                    thread_pool_mutex.lock().unwrap().execute(|| udp::handler(req));
                }
            }),
        });
    }

    for thread in threads {
        thread.join().unwrap();
    }
}
